# Build Brick Row - Liviu-George Bitiusca 2016
import maya.cmds as cmds
import math

# class for creating a row of bricks of variable size and with variable face extrusion
# supports dynamic member (brick) deletion using an index
# supports manipulation and placement between multiple rows of bricks, even after creation
# (e.g. rows can be moved on top of other rows, bricks can be deleted)

# the script was used to generate a few rows of bricks of some of the buildings in the
# associated CGI Tools animation project.

# NOTE: there are no requirements or conditions to be met for the script to work.
# Simply run the class and then use its methods to start building geometry
# There are a few lines of usage example at the bottom of the file

class BrickRow:
    def __init__(self, copies, rowWidth, y, z,
        facingAxis = 'z', extrudeDepth = 0, extrudeOffset = 0):
        '''
        copies (int) | number of bricks
        rowWidth (float) | width of the row
        y, z (float) | height, depth of the brick
        facingAxis (string) | axis that the bricks face towards (x/-x/z/-z)
        extrudeDepth | extrude depth
        extrudeOffset (float) | extrude offset
        ---
        returns name of created brick group (string)
        '''
        # param error checking
        if copies < 1:
            cmds.warning('You need to create at least 1 brick copy. Nothing created.')
            return
        if rowWidth <= 0 or y <= 0 or y <= 0:
            cmds.error('Invalid row or brick dimensions (one or more sides less than 0).')
        if facingAxis not in ('x','-x','z','-z'):
            cmds.error('BrickRow only works with x/-x/z/-z axes.')
        
        # Instance variables
        self.brickList = [0]*copies
        self.brickCount = copies
        self.name = 'BrickGroup'
        self.extrudeNode = 'none'
        self.extrudeDepth = 0
        self.extrudeOffset = 0
        self.translation = [0, 0, 0]
        self.rotation = [0, 0, 0]
        self.width = rowWidth
        self.height = y
        self.depth = z
        
        # local variables
        brickWidth = rowWidth / copies
        
        # Create base brick
        brick = cmds.polyCube(name = 'brick#', width = brickWidth, height = y, depth = z)[0];
        
        # Store first brick's name
        self.brickList[0] = brick
        
        # Extrude
        if extrudeDepth:
            cmds.select('{0}.f[0]'.format(brick), replace = True)
            self.extrudeNode = cmds.polyExtrudeFacet(offset = extrudeOffset,
                thickness = extrudeDepth, smoothingAngle = 0)[0]
            self.extrudeDepth = extrudeDepth
            self.extrudeOffset = extrudeOffset
            
        # Duplicate brick
        for i in range(1, copies):
            brick = cmds.duplicate(brick, inputConnections = True)[0]
            cmds.move(brickWidth, 0, 0, brick, relative = True)
            
            # Add to list
            self.brickList[i] = brick
        
        # Create group from list
        self.name = cmds.group(self.brickList, n = 'BrickGroup#')
        
        # center group along the x-axis
        cmds.move(-rowWidth/2 + brickWidth/2, 0, 0, self.name)
        
        # move upwards to align row bottom to the grid, align pivot
        cmds.move(0, self.height/2, 0, self.name, relative = True)
        cmds.move(0, 0, 0, self.name + '.scalePivot')
        cmds.move(0, 0, 0, self.name + '.rotatePivot')
        
        # freeze translation
        cmds.makeIdentity(self.name, apply = True, t = True)
        
        # rotate if needed
        if facingAxis == '-z':
            cmds.rotate(0, 180, 0, self.name)
            self.rotation[1] = 180
        if facingAxis == 'x':
            cmds.rotate(0, 90, 0, self.name)
            self.rotation[1] = 90
        if facingAxis == '-x':
            cmds.rotate(0, -90, 0, self.name)
            self.rotation[1] = -90
        
    def deleteBricks(self, bricks = []):
        '''
        bricks[] (int array) | indices of bricks to delete
        fromCenter (bool) | enable deletion form center
        '''
        # Delete bricks
        for i in bricks:
            if self.brickList[i] == 0:
                print 'Brick', i, 'already deleted'
            else:
                cmds.delete(self.brickList[i])
                self.brickList[i] = 0
        return self
                
    def deleteFromBrick(self, center = 'center', left = 0, right = 'NULL'):
        '''
        center (int) | central brick  from where to start deleting (going left and right)
        left (int) | number of bricks to delete from left of center
        right (int) | number of bricks to delete from right of center
        '''
        # check for center keyword
        if center == 'center':
            center = int( math.floor( (len(self.brickList) - 1) / 2 ) )
        
        # check if center, left and/or right are not out of bounds
        if center < 0 or center > len(self.brickList) - 1:
            cmds.error('"center" index out of valid range.')
               
        # if left or right are negative, just set them to 0
        if left < 0:
            left = 0
        if right < 0:
            right = 0
            
        # set left = right if only one bound has been specified
        if right == 'NULL':
            right = left
            
        # check if bounds don't go outside valid range
        leftBound = center - left
        if leftBound < 0:
            leftBound = 0
            
        rightBound = center + right
        if rightBound > len(self.brickList) - 1:
            rightBound = len(self.brickList) - 1
        
        # loop within range, delete everything
        for i in range(leftBound, rightBound + 1):
            if self.brickList[i] == 0:
                print 'Brick', i, 'already deleted. Skipped.'
            else:
                cmds.delete(self.brickList[i])
                self.brickList[i] = 0
                
        # return self for method chaining
        return self
        
    def move(self, x, y, z, relative = False):
        '''
        x, y, z (float) | amounts to move in x, y, z
        relative (bool) | enable relative move
        '''
        cmds.move(x, y, z, self.name, relative = relative)
        if relative:
            self.translation[0] += x
            self.translation[1] += y
            self.translation[2] += z
        else:
            self.translation = [x, y, z]
        
        # return self for method chaining
        return self
    
    def stackOn(self, row):
        # if object doesn't exist, throw error
        if not cmds.objExists(row.name):
            cmds.error('Specified object (', row , ') does not exist.')
        
        # get translation, rotation and height of the row to stack on
        translation = row.translation
        rotation = row.rotation
        height = row.height
              
        # match its translation and rotation
        newT = [translation[0], translation[1] + height, translation[2]]
        cmds.move(newT[0], newT[1], newT[2], self.name)
        cmds.rotate(rotation[0], rotation[1], rotation[2], self.name)
        
        # store new translation values
        self.translation = newT
        self.rotation = rotation
        
        # return self for method chaining
        return self
        
    def extrude(self, d = 'null', o = 'null', r = False):
        '''
        ATTENTION: undo will work, but will not modify the internal values of
        extrudeDepth and extrudeOffset. Avoid undoing if possible.
        
        d (float) | extrude depth (thickness)
        o (float) | extrude offset
        r (bool) | relative extrusion
        '''
        
        # if no node stored on the node, return
        if self.extrudeNode == 'none':
            cmds.warning('Input row has no brick extrusion. Nothing changed.')
            return self
        
        # else modify the brick extrusion
        if not r:
            if d != 'null':
                cmds.setAttr(self.extrudeNode + '.thickness', d)
                self.extrudeDepth = d
            
            if o != 'null':
                cmds.setAttr(self.extrudeNode + '.offset', o)
                self.extrudeOffset = o
        else:
            if d != 'null':
                cmds.setAttr(self.extrudeNode + '.thickness', self.extrudeDepth + d)
                self.extrudeDepth += d
                
            if o != 'null':
                cmds.setAttr(self.extrudeNode + '.offset', self.extrudeOffset + o)
                self.extrudeOffset += o
        
        # return self for method chaining
        return self           


### Example of use ###
# actual code used in generating the modelling elements in the
# associated animation project

# set a width for the wall
buildingWidth = 3.71

# create first row of bricks...
# ...and delete 1 brick left and right starting from the center     
row = BrickRow(copies = 13, rowWidth = buildingWidth, y = 0.2, z = 0.2,
    facingAxis = 'x', extrudeDepth = 0.01, extrudeOffset = 0.02)
row.deleteFromBrick('center', 1)

# create second row of bricks...
# ...stack it on top of the first row
# ...then delete 2 bricks on the left and 3 on the right starting from center
row2 = BrickRow(copies = 26, rowWidth = buildingWidth, y = 0.1, z = 0.2,
    facingAxis = 'x', extrudeDepth = 0.01, extrudeOffset = 0.02)
row2.stackOn(row).deleteFromBrick('center', 2, 3)

# crate third row of bricks...
# ...stack it on second row
# ...delete 1 brick left and right from the center
# ...and alter the extrude of the bricks
row3 = BrickRow(copies = 13, rowWidth = buildingWidth, y = 0.05, z = 0.2,
    facingAxis = 'x', extrudeDepth = 0.01, extrudeOffset = 0.01)
row3.stackOn(row2).deleteFromBrick('center', 1)
row3.extrude(d = 0.01, r = True)
